/*
 * libxml-dtc provides an XML framework creation facilities
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#define _GNU_SOURCE
#include <sitemap.h>
#include <string.h>

int xml_dtc_sitemap_add(xml_dtc_sitemap *sitemap, xml_dtc_object *object) {
  if (sitemap && object) {
    xml_dtc_hash *container = NULL;
    xml_dtc_type *pipeline = xml_dtc_type_get("pipeline");
    container = (xml_dtc_hash *)xml_dtc_hash_get(sitemap->hash, object->type->name);
    if (container == NULL) {
      /* Transitional hack */
      if (object->type == pipeline) {
	container = (xml_dtc_hash *)xml_dtc_object_list_new(sitemap->resource);
	xml_dtc_hash_add_cast(sitemap->hash, "pipeline", container);
	xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Adding pipeline to sitemap");
      } else {
	container = xml_dtc_object_hash_new(sitemap->resource);
	xml_dtc_hash_add_cast(sitemap->hash, object->type->name, container);
	xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Adding '%s' to sitemap", object->type->name);
      }
    }
    if (object->type == pipeline) {
      return xml_dtc_list_push_cast((xml_dtc_list *)container, object);
    } else {
      return xml_dtc_object_hash_add(container, object);
    }
  } else {
    return -1;
  }
}

xml_dtc_object *xml_dtc_sitemap_get(xml_dtc_sitemap *sitemap, char *type, char *name) {
  return xml_dtc_hash_get(xml_dtc_sitemap_get_all_type(sitemap, type), name);
}

xml_dtc_hash *xml_dtc_sitemap_get_all_type(xml_dtc_sitemap *sitemap, char *type) {
  if (sitemap)
    if (type)
      return (xml_dtc_hash *)xml_dtc_hash_get(sitemap->hash, type);
  return NULL;
}

int xml_dtc_sitemap_set_preferred_by_key(xml_dtc_sitemap *sitemap, char *type, char *name) {
  if (sitemap && type && name)
    return xml_dtc_hash_set_default((xml_dtc_hash *)xml_dtc_hash_get(sitemap->hash, type), name);
  return -1;
}

int xml_dtc_sitemap_free(xml_dtc_sitemap *sitemap) {
  if (sitemap == NULL)
    return -1;

  if (sitemap->resource)
    xml_dtc_resource_unlink(sitemap->resource);

  if (sitemap->base_dir)
    free(sitemap->base_dir);

  if (sitemap->hash) {
    int i;
    for (i = 0; i < sitemap->hash->num; i++) {
      if (strncmp(sitemap->hash->list[i].key, "pipeline", 9) == 0) {
	xml_dtc_list_free((xml_dtc_list *)sitemap->hash->list[i].element);
	free(sitemap->hash->list[i].key);
	sitemap->hash->list[i].key = NULL;
	sitemap->hash->list[i].element = NULL;
      }
    }
    xml_dtc_hash_free(sitemap->hash);
  }

  free(sitemap);

  return 0;
}

xml_dtc_sitemap *xml_dtc_sitemap_new(const xml_dtc_resource *resource) {
  xml_dtc_sitemap *result;

  if (resource == NULL)
    return NULL;

  result = (xml_dtc_sitemap *)malloc(sizeof(xml_dtc_sitemap));
  memset(result, 0, sizeof(xml_dtc_sitemap));
  result->resource = xml_dtc_resource_link(resource);

  result->hash = xml_dtc_hash_new_cast(resource, free, strdup, xml_dtc_hash_free, NULL);

  return result;
}

xml_dtc_sitemap *xml_dtc_sitemap_new_from_file(xml_dtc_resource *resource, char *file) {
  xml_dtc_sitemap *result = NULL;
  xml_dtc_handler *handler;
  xmlDocPtr xmldoc;
  xmlNodePtr node;
  char *tmp;
  int i;
  xml_dtc_msg(resource, XML_DTC_MSG_DEBUG, "Opening file '%s'.", file);

  xmldoc = xmlParseFile(file);

  if (xmldoc == NULL) {
    xml_dtc_msg(resource, XML_DTC_MSG_ERROR, "File '%s' not parsed successfully.", file);
    return NULL;
  }

  node = xmlDocGetRootElement(xmldoc);
  if (node == NULL) {
    xml_dtc_msg(resource, XML_DTC_MSG_ERROR, "File '%s' does not have a root element.", file);
    xmlFreeDoc(xmldoc);
    return NULL;
  }

  if (xmlStrcmp(node->ns->href, XML_DTC_SITEMAP_NAMESPACE)
      || xmlStrcmp(node->name, (const xmlChar *) XML_DTC_SITEMAP_ROOT_NODE)) {
    xml_dtc_msg(resource, XML_DTC_MSG_ERROR, "File doesn't have the right root node.");
    xmlFreeDoc(xmldoc);
    return NULL;
  }
  
  result = xml_dtc_sitemap_new(resource);
  
  if (xml_dtc.initializers)
    for (i = 0; i < xml_dtc.initializers->len; i++) {
      if (((xml_dtc_handler *)xml_dtc.initializers->list[i])->handle((xml_dtc_handler *)xml_dtc.initializers->list[i], xml_dtc.handlers, result, node) == NULL) {
	xml_dtc_msg(result->resource, XML_DTC_MSG_ERROR, "Error while initializing '%s' sitemap.", file);
	xml_dtc_sitemap_free(result);
	xmlFreeDoc(xmldoc);
	return NULL;
      }
    }

  handler = (xml_dtc_handler *)xml_dtc_hash_get((xml_dtc_hash *)xml_dtc_hash_get(xml_dtc.handlers, "xml-dtc"), "sitemap");

  if (handler) {
    if (handler->handle(handler, xml_dtc.handlers, result, node) == NULL) {
      xml_dtc_msg(result->resource, XML_DTC_MSG_ERROR, "Error while reading '%s' sitemap.", file);
      xml_dtc_sitemap_free(result);
      xmlFreeDoc(xmldoc);
      return NULL;
    }
  } else {
    xml_dtc_sitemap_free(result);
    xmlFreeDoc(xmldoc);
    return NULL;
  }
  
  xmlFreeDoc(xmldoc);

  if ((tmp = rindex(file, '/'))) {
    result->base_dir = strndup(file, (int)tmp - (int)file);
  }
#if 0
  return NULL;
#endif
  return result;
}

int xml_dtc_sitemap_request(xml_dtc_sitemap *sitemap, xml_dtc_request *request) {
  if (!xml_dtc.resource)
    return -1;
  if (sitemap && request) {
    xml_dtc_runner *runner;
    runner = (xml_dtc_runner *)xml_dtc.runners->list[xml_dtc.runners->len - 1];
    request->sitemap = sitemap;
    if (runner->run(request) == 0) {
      if (request->error.type && ! request->result) {
	return xml_dtc_request_serialize_error(request);
      }
      return 0;
    }
  }
  return -1;
}

int xml_dtc_sitemap_resolve_relative_path(xml_dtc_sitemap *sitemap, char **path) {
  if (sitemap && path && *path) {
    if ((*path[0] != '/') && sitemap->base_dir) {
      char *tmp;
      tmp = malloc(strlen(*path) + strlen(sitemap->base_dir) + 2);
      sprintf(tmp, "%s/%s", sitemap->base_dir, *path);
      free(*path);
      *path = tmp;
      return 0;
    }
    return 0;
  }
  return -1;
}
