/*
 * libxml-dtc provides an XML framework creation facilities
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#define _GNU_SOURCE

#include <modules.h>
#include <xml-dtc.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <dirent.h>
#include <config.h>

#ifdef MODULES_DIR
#include <ltdl.h>
#endif

#include "static.h"

struct _xml_dtc xml_dtc = {
  NULL, NULL,
  NULL, NULL, NULL, NULL, NULL
};

struct _xml_dtc_types xml_dtc_types = {
  NULL,
  0, 0, NULL
};

#define SIZE_MOD 64

int xml_dtc_cleanup() {
  if (xml_dtc.handlers)
    xml_dtc_hash_free(xml_dtc.handlers);
  if (xml_dtc.runners)
    xml_dtc_list_free(xml_dtc.runners);
  if (xml_dtc.initializers)
    xml_dtc_list_free(xml_dtc.initializers);
  if (xml_dtc.hooks)
    xml_dtc_list_free(xml_dtc.hooks);
  if (xml_dtc.contexts)
    xml_dtc_hash_free(xml_dtc.contexts);
  if (xml_dtc.modules) {
    int i;
    for (i = 0; i < xml_dtc.modules->num; i++) {
      if (((xml_dtc_module *)xml_dtc.modules->list[i].element)->close)
	((xml_dtc_module *)xml_dtc.modules->list[i].element)->close();
#ifdef MODULES_DIR
      if (((xml_dtc_module *)xml_dtc.modules->list[i].element)->_private)
	lt_dlclose(((xml_dtc_module *)xml_dtc.modules->list[i].element)->_private);
#endif
    }
    xml_dtc_hash_free(xml_dtc.modules);
  }
  if (xml_dtc.resource)
    xml_dtc_resource_unlink(xml_dtc.resource);
  if (xml_dtc_types.resource)
    xml_dtc_resource_unlink(xml_dtc_types.resource);
#ifndef OBJECT_DEBUG
  if (xml_dtc_types.size) {
    int i;
    for (i = 0; i < xml_dtc_types.num; i++)
      if (xml_dtc_types.hash[i].name) 
	free(xml_dtc_types.hash[i].name);
    free(xml_dtc_types.hash);
    memset(&xml_dtc_types, 0, sizeof(xml_dtc_types));
  }
#endif
  memset(&xml_dtc, 0, sizeof(xml_dtc));
  return 0;
}

#ifdef MODULES_DIR
static int xml_dtc_load_module(char *filename) {
  lt_dlhandle dl;
  xml_dtc_module *module;

  xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_NOTICE, "Loading %s module.", filename);

  /* TODO: Handle errors */
  dl = lt_dlopen(filename);
  if (!dl) {
    xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_ERROR, "%s",lt_dlerror());
    return -1;
  } else {
    const char *error;
    char *module_name, *tmp;

    module_name = rindex(filename, '/');
    if (module_name) {
      module_name++;
    } else {
      module_name = filename;
    }
    module_name[strlen(module_name) - 3] = 0;
    tmp = malloc(16 + strlen(module_name));
    sprintf(tmp, "xml_dtc_%s_module", module_name);
    module_name[strlen(module_name) - 3] = '.';
    module = lt_dlsym(dl, tmp);
    free(tmp);
    if ((error = lt_dlerror()) != NULL) {
      xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_ERROR, "%s",error);
      return -1;
    } else {
      if (module->API == XML_DTC_API_VERSION) {
	if (xml_dtc_object_hash_add(xml_dtc.modules, module) == 0) {
	  module->registr();
	  module->_private = dl;
	} else
	  xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_ERROR, "Module \"%s\" already loaded.", module->name);
      } else {
	xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_ERROR, "API version mismatch.");
	return -1;
      }
    }
  }

  return 0;
}

static int xml_dtc_recursively_load_modules(char *path) {
  DIR *dir;
  struct dirent *current;

  xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_NOTICE, "Loading modules from %s.", path);

  dir = opendir(path);

  if (dir) {
    while ((current = readdir(dir))) {
      if (current->d_type == DT_DIR) {
	if (! ((current->d_name[0] == '.') && ((current->d_name[1] == 0) || ((current->d_name[1] == '.') && (current->d_name[2] == 0))))) {
	  char *new_dir;
	  
	  new_dir = malloc(strlen(path) + strlen(current->d_name) + 2);
	  strncpy(new_dir, path, strlen(path));
	  new_dir[strlen(path)] = '/';
	  strncpy(new_dir + strlen(path) + 1, current->d_name, strlen(current->d_name) + 1);
	  
	  xml_dtc_recursively_load_modules(new_dir);
	  
	  free(new_dir);	  
	}
      } else if (current->d_type == DT_REG) {
	int len;
	if (((len = strlen(current->d_name)) > 3) && (current->d_name[len - 1] == 'a') && (current->d_name[len - 2] == 'l') && (current->d_name[len - 3] == '.')) {
	  char *new_dir;
	  
	  new_dir = malloc(strlen(path) + strlen(current->d_name) + 2);
	  strncpy(new_dir, path, strlen(path));
	  new_dir[strlen(path)] = '/';
	  strncpy(new_dir + strlen(path) + 1, current->d_name, strlen(current->d_name) + 1);
	  
	  xml_dtc_load_module(new_dir);

	  free(new_dir);	  
	}
      }
    }
    closedir(dir);
    return 0;
  }
  return -1;
}
#endif

int xml_dtc_init_modules() {
  int i = 0;
  xml_dtc_module *module;

  while ((module = xml_dtc_static_modules[i])) {
    xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_NOTICE, "Initializing internal module \"%s\"", module->name);

    if (xml_dtc.modules == NULL)
      xml_dtc.modules = xml_dtc_hash_new(xml_dtc.resource, NULL, NULL, NULL, NULL);

    xml_dtc_object_hash_add(xml_dtc.modules, (xml_dtc_object *)module);
    module->registr();
    module->_private = NULL;
    i++;
  }

#ifdef MODULES_DIR
  if (lt_dlinit() == 0) {
    char *path;
    path = getenv("XML_DTC_MODULES_DIR");
    if (path == NULL)
      path = MODULES_DIR;

    xml_dtc_recursively_load_modules(path);
  }
#endif

  if (!xml_dtc.modules) {
    xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_ERROR, "Cannot initialize " PACKAGE ".");
    xml_dtc_cleanup();
    return -1;
  }
  return 0;
}

int _xml_dtc_caller_free(xml_dtc_object *list);

int xml_dtc_init(const xml_dtc_resource *resource) {
  if (xml_dtc.resource)
    return -1;
  
  if (! resource)
    resource = xml_dtc_resource_new();
  
  //    xml_dtc_resource_set_verbosity(resource, XML_DTC_MSG_DEBUG);
  
  xml_dtc.resource = xml_dtc_resource_link(resource);
  xml_dtc_types.resource = xml_dtc_resource_link(resource);
  xml_dtc_register_type("component", sizeof(xml_dtc_component), NULL, NULL);
  xml_dtc_register_type("caller", sizeof(xml_dtc_caller), _xml_dtc_caller_free, NULL);

  return 0;
}

int xml_dtc_handler_free(xml_dtc_handler *handler) {
  if (handler) {
    if (handler->name)
      free(handler->name);
    if (handler->parent_node)
      free(handler->parent_node);
    free(handler);
    return 0;
  }
  return -1;
}

int xml_dtc_runner_free(xml_dtc_runner *runner) {
  if (runner) {
    if (runner->name)
      free(runner->name);
    if (runner->parent_node)
      free(runner->parent_node);
    free(runner);
    return 0;
  }
  return -1;
}

int xml_dtc_register_initializer(const char *name, xml_dtc_handler_callback callback, void *private_data) {
  xml_dtc_handler *result;
  result = (xml_dtc_handler *)malloc(sizeof(xml_dtc_handler));
  memset(result, 0, sizeof(xml_dtc_handler));
  result->name = strdup(name);
  result->handle = callback;
  result->_private = private_data;

  if (! xml_dtc.initializers) {
    xml_dtc.initializers = xml_dtc_list_new_cast(xml_dtc.resource, xml_dtc_handler_free, NULL);
  }
  xml_dtc_list_push_cast(xml_dtc.initializers, result);

  return 0;
}

int xml_dtc_register_handler(const char *name, const char *parent, xml_dtc_handler_callback callback, void *private_data) {
  xml_dtc_handler *result;
  xml_dtc_hash *tmp;

  xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_NOTICE, "Registering %s handler, with %s parent.", name, parent);
  result = (xml_dtc_handler *)malloc(sizeof(xml_dtc_handler));
  memset(result, 0, sizeof(xml_dtc_handler));
  result->name = strdup(name);
  result->parent_node = strdup(parent);
  result->handle = callback;
  result->_private = private_data;

  if (! xml_dtc.handlers) {
    xml_dtc.handlers = xml_dtc_hash_new_cast(xml_dtc.resource, free, strdup, xml_dtc_hash_free, NULL);
  }

  if (! (tmp = (xml_dtc_hash *)xml_dtc_hash_get(xml_dtc.handlers, parent))) {
    tmp = xml_dtc_hash_new_cast(xml_dtc.resource, free, strdup, xml_dtc_handler_free, NULL);
    xml_dtc_hash_add(xml_dtc.handlers, (char *)parent, tmp);
  }
  xml_dtc_hash_add(tmp, result->name, result);

  return 0;
}

int xml_dtc_register_runner(const char *name, xml_dtc_runner_callback callback, void *private_data) {
  xml_dtc_runner *result;
  xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_NOTICE, "Registering %s runner.", name);
  result = (xml_dtc_runner *)malloc(sizeof(xml_dtc_runner));
  memset(result, 0, sizeof(xml_dtc_runner));
  result->name = strdup(name);
  result->run = callback;
  result->_private = private_data;

  if (! xml_dtc.runners) {
    xml_dtc.runners = xml_dtc_list_new_cast(xml_dtc.resource, xml_dtc_runner_free, NULL);
  }
  xml_dtc_list_push_cast(xml_dtc.runners, result);

  return 0;
}

int xml_dtc_register_hook(xml_dtc_hook callback) {
  xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_NOTICE, "Registering a hook.");
  if (! xml_dtc.hooks) {
    xml_dtc.hooks = xml_dtc_value_list_new(xml_dtc.resource);
  }
  xml_dtc_list_push_cast(xml_dtc.hooks, callback);

  return 0;
}

xml_dtc_type *xml_dtc_type_get(const char *name) {
  if (name && xml_dtc_types.num) {
    int i;
    for (i = 0; i < xml_dtc_types.num; i++)
      if (strcmp(name, xml_dtc_types.hash[i].name) == 0)
	return &xml_dtc_types.hash[i];
  }
  return NULL;
}

int xml_dtc_register_type(const char *name, int size, xml_dtc_object_free_callback free, const char *parent) {
  if (name && size) {
    if (xml_dtc_type_get(name)) {
      xml_dtc_msg(xml_dtc_types.resource, XML_DTC_MSG_ERROR, "Cannot register type '%s': already registered.", name);
      return -1;
    }
    if (xml_dtc_types.size <= xml_dtc_types.num) {
      xml_dtc_types.hash = realloc(xml_dtc_types.hash, sizeof(xml_dtc_type) * (xml_dtc_types.size + SIZE_MOD));
      xml_dtc_types.size += SIZE_MOD;
    }
    if (parent) {
      xml_dtc_type *tmp = xml_dtc_type_get(parent);
      if (tmp) {
	if (tmp->size > size) {
	  xml_dtc_msg(xml_dtc_types.resource, XML_DTC_MSG_ERROR, "Cannot register type '%s': parent type '%s' is bigger (%d vs. %d).", name, parent, tmp->size, size);
	  return -1;
	}
	xml_dtc_types.hash[xml_dtc_types.num].parent = xml_dtc_type_get(parent);
      }
      else {
	xml_dtc_msg(xml_dtc_types.resource, XML_DTC_MSG_ERROR, "Cannot register type '%s': parent type '%s' is not registered.", name, parent);
	return -1;
      }
    } else {
      xml_dtc_types.hash[xml_dtc_types.num].parent = NULL;
    }
    xml_dtc_types.hash[xml_dtc_types.num].name = strdup(name);
    xml_dtc_types.hash[xml_dtc_types.num].size = size;
    xml_dtc_types.hash[xml_dtc_types.num++].free = free;
    if (parent)
      xml_dtc_msg(xml_dtc_types.resource, XML_DTC_MSG_DEBUG, "Registering '%s' type of size %d with '%s' parent.", name, size, parent);
    else
      xml_dtc_msg(xml_dtc_types.resource, XML_DTC_MSG_DEBUG, "Registering '%s' type of size %d.", name, size);
  }
  return 0;
}
