/*
 * This module provides actions to the XML-DTC framework
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This module is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this module; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include <modules.h>
#include <xml-dtc.h>
#include <mappings.h>
#include <list.h>
#include <string.h>

typedef struct _xml_dtc_action_caller xml_dtc_action_caller;

typedef struct _xml_dtc_action xml_dtc_action;
struct _xml_dtc_action {
  XML_DTC_COMPONENT_STANDARD_DEF;
};

struct _xml_dtc_action_caller {
  XML_DTC_CALLER_STANDARD_DEF;
  xml_dtc_list *sub_pipeline;
};

static int _xml_dtc_action_caller_free(xml_dtc_action_caller *caller) {
  if (caller->sub_pipeline)
    xml_dtc_list_free(caller->sub_pipeline);
  return 0;
}

static xml_dtc_object *xml_dtc_handler_action_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xmlChar *attr;
  xml_dtc_action *result;

  attr = xmlGetProp(node, "name");
  result = (xml_dtc_action *)xml_dtc_object_new(sitemap->resource, attr, (char *)node->name);
  result->handle = handler->_private;
  free(attr);

  xml_dtc_sitemap_add(sitemap, (xml_dtc_object *) result);
  return (xml_dtc_object *)handler;
}

static xml_dtc_object *xml_dtc_handler_act_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xml_dtc_action_caller *result;
  xml_dtc_hash *good_handlers;

  result = (xml_dtc_action_caller *)xml_dtc_object_new(sitemap->resource, "action", "act-caller");
  result->component.name = xmlGetProp(node, "type");

  good_handlers = (xml_dtc_hash *) xml_dtc_hash_get(handlers, "pipeline");

  result->sub_pipeline = xml_dtc_object_list_new(sitemap->resource);

  node = node->children;

  while (node) {
    /* Ignore Blank Nodes */
    if (! xmlIsBlankNode(node)) {
      /* Test if node is an element or something else (PI, Comment...) */
      if (node->type != XML_ELEMENT_NODE) {
        /* Comments are allowed */
        if (node->type == XML_COMMENT_NODE) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Found comment");
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Element or comment expected");
          return NULL;
        }
      } else {
        /* Check if namespace is good */
        if ((node->ns != node->parent->ns) && (xmlStrcmp(node->ns->href, XML_DTC_SITEMAP_NAMESPACE) != 0)) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "expecting %s namespace, but got %s", XML_DTC_SITEMAP_NAMESPACE,
                      (node->ns != NULL) ? (char *) node->ns->href : "none");
          return NULL;
        }
        /* Now we know that namespace is ok, we can check and initialize all stuff */
        /* Search a handler for this node */
	if (xmlStrcmp(node->name, "parameter") == 0) {
	  xmlChar *tmp1, *tmp2;
	  if (!result->params) {
	    result->params = xml_dtc_string_hash_new(result->resource);
	  }
	  xml_dtc_string_hash_add(result->params, tmp1 = xmlGetProp(node, "name"), tmp2 = xmlGetProp(node, "value"));
	  xmlFree(tmp1);
	  xmlFree(tmp2);
	} else {
          xml_dtc_handler *tmp;
          xml_dtc_object *object;
          if ((tmp = (xml_dtc_handler *)xml_dtc_hash_get(good_handlers, (char *)node->name))) {
            xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Found handler for %s", node->name);
            if ((object = tmp->handle(tmp, handlers, sitemap, node)) == NULL)
              return NULL;
            else
              xml_dtc_list_push_cast(result->sub_pipeline, object);
          } else {
            xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "No handler for %s", node->name);
            return NULL;
          }
        }
      }
    }
    node = node->next;
  }

  return (xml_dtc_object *)result;
}

static int xml_dtc_act_http_header(xml_dtc_action_caller *caller, xml_dtc_request *request) {
  int i;
  char *key, *value;

  if (caller->params) {
    for(i = 0; i < caller->params->num; i++) {
      key = xml_dtc_resolve_mappings(request->mappings, caller->params->list[i].key);
      value = xml_dtc_resolve_mappings(request->mappings, (char *)caller->params->list[i].element);
      
      xml_dtc_request_set_out_header(request, key, value);
      free(value);
      free(key);
    }
  }
  xml_dtc_mappings_new_list(request->mappings);
  i = xml_dtc_request_apply(request, caller->sub_pipeline);
  xml_dtc_mappings_destroy_last(request->mappings);
  return i;
}

static int xml_dtc_act_set_cookie(xml_dtc_action_caller *caller, xml_dtc_request *request) {
  int i;
  char *key, *value;

  if (caller->params) {
    for(i = 0; i < caller->params->num; i++) {
      key = xml_dtc_resolve_mappings(request->mappings, caller->params->list[i].key);
      value = xml_dtc_resolve_mappings(request->mappings, (char *)caller->params->list[i].element);
      
      xml_dtc_request_set_cookie(request, key, value);
      free(value);
      free(key);
    }
  }
  xml_dtc_mappings_new_list(request->mappings);
  i = xml_dtc_request_apply(request, caller->sub_pipeline);
  xml_dtc_mappings_destroy_last(request->mappings);
  return i;
}

static int xml_dtc_actions_module_register(void) {
  xml_dtc_register_handler("actions", "components", (xml_dtc_handler_callback) xml_dtc_handler_components_loop_callback, NULL);
  xml_dtc_register_handler("act", "pipeline", (xml_dtc_handler_callback) xml_dtc_handler_act_callback, NULL);
  xml_dtc_register_handler("http_header", "actions", (xml_dtc_handler_callback) xml_dtc_handler_action_callback, xml_dtc_act_http_header);
  xml_dtc_register_handler("set_cookie", "actions", (xml_dtc_handler_callback) xml_dtc_handler_action_callback, xml_dtc_act_set_cookie);
  xml_dtc_register_type("action", sizeof(xml_dtc_action), NULL, "component");
  xml_dtc_register_type("act-caller", sizeof(xml_dtc_action_caller), (xml_dtc_object_free_callback) _xml_dtc_action_caller_free, "caller");
  return 0;
}

xml_dtc_module xml_dtc_actions_module = {
  XML_DTC_MODULE_STANDARD_STUFF,
  xml_dtc_actions_module_register
};
