/*
 * This module provides aggregation to the XML-DTC framework
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This module is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this module; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include <modules.h>
#include <xml-dtc.h>
#include <mappings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>

typedef struct _xml_dtc_aggregator_caller xml_dtc_aggregator_caller;

typedef struct _xml_dtc_aggregate_part xml_dtc_aggregate_part;

typedef struct _xml_dtc_aggregator xml_dtc_aggregator;
struct _xml_dtc_aggregator {
  XML_DTC_COMPONENT_STANDARD_DEF;
};

struct _xml_dtc_aggregator_caller {
  XML_DTC_CALLER_STANDARD_DEF;
  xmlChar *element, *ns, *prefix;
  xml_dtc_list *parts;
};

struct _xml_dtc_aggregate_part {
  XML_DTC_OBJECT_STANDARD_DEF;
  xmlChar *element, *ns, *prefix, *src;
  int strip_root;
};

static int _xml_dtc_aggregator_caller_free(xml_dtc_aggregator_caller *caller) {
  if (caller->element)
    xmlFree(caller->element);
  if (caller->ns)
    xmlFree(caller->ns);
  if (caller->prefix)
    xmlFree(caller->prefix);
  if (caller->parts)
    xml_dtc_list_free(caller->parts);
  return 0;
}

static int _xml_dtc_aggregate_part_free(xml_dtc_aggregate_part *part) {
  if (part->element)
    xmlFree(part->element);
  if (part->ns)
    xmlFree(part->ns);
  if (part->prefix)
    xmlFree(part->prefix);
  if (part->src)
    xmlFree(part->src);
  return 0;
}

static int xml_dtc_aggregate(xml_dtc_aggregator_caller *caller, xml_dtc_request *request) {
  xmlDocPtr xmldoc;
  xmlNodePtr root_node;
  xmlNsPtr ns = NULL;
  int i;

  xmldoc = xmlNewDoc("1.0");

  if (xmldoc == NULL) {
    xml_dtc_msg(caller->resource, XML_DTC_MSG_ERROR, "Could not create XML.");
    return -1;
  }

  if (caller->ns)
    ns = xmlNewNs(NULL, caller->ns, caller->prefix);

  root_node = xmlNewDocNode(xmldoc, ns, caller->element, NULL);
  xmlDocSetRootElement(xmldoc, root_node);
  root_node->nsDef=ns;

  for (i = 0; i < caller->parts->len; i++) {
    xmlNodePtr node, partroot;
    xmlDocPtr partdoc;
    xml_dtc_aggregate_part *part;
    char *filename;
    part = (xml_dtc_aggregate_part *)caller->parts->list[i];
    filename = xml_dtc_resolve_mappings(request->mappings, part->src);
    xml_dtc_sitemap_resolve_relative_path(request->sitemap, &filename);
    if (filename) {
      partdoc = xmlParseFile(filename);
      free(filename);
      partroot = xmlDocGetRootElement(partdoc);
      if (part->strip_root) {
	partroot = partroot->children;
	while (partroot) {
	  node = xmlNewChild(root_node, NULL, BAD_CAST "null", NULL);
	  xmlReplaceNode(node, xmlDocCopyNode(partroot, xmldoc, 1));
	  partroot = partroot->next;
	}
      } else {
	node = xmlNewChild(root_node, NULL, BAD_CAST "null", NULL);
	xmlReplaceNode(node, xmlDocCopyNode(partroot, xmldoc, 1));
      }
    /*    if (part->ns)
	  ns = xmlNewNs(NULL, part->ns, part->prefix);*/
    }
  }

  request->xmltree = xmldoc;

  return 0;
}

static xml_dtc_object *xml_dtc_handler_aggregator_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xml_dtc_aggregator *result;

  result = (xml_dtc_aggregator *)xml_dtc_object_new(sitemap->resource, "aggregator", "aggregator");
  result->handle = (int(*)(xml_dtc_caller *, xml_dtc_request *))xml_dtc_aggregate;

  xml_dtc_sitemap_add(sitemap, (xml_dtc_object *) result);
  xml_dtc_sitemap_set_preferred_by_key(sitemap, "aggregator", "aggregator");

  return (xml_dtc_object *)handler;
}

static xml_dtc_object *xml_dtc_handler_aggregate_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xml_dtc_aggregator_caller *result;
  result = (xml_dtc_aggregator_caller *)xml_dtc_object_new(sitemap->resource, "aggregator", "aggregate-caller");
  result->element = xmlGetProp(node, "element");
  result->ns = xmlGetProp(node, "ns");
  result->prefix = xmlGetProp(node, "prefix");

  result->parts = xml_dtc_object_list_new(sitemap->resource);

  node = node->children;

  while(node) {
    xml_dtc_aggregate_part *part;
    xmlChar *tmp;

    part = (xml_dtc_aggregate_part *)xml_dtc_object_new(sitemap->resource, "aggregate-part", "aggregate-part");

    part->src = xmlGetProp(node, "src");
    part->element = xmlGetProp(node, "element");
    part->ns = xmlGetProp(node, "ns");
    part->prefix = xmlGetProp(node, "prefix");
    part->strip_root = xmlStrEqual(tmp = xmlGetProp(node, "strip-root"), "true");
    xmlFree(tmp);

    xml_dtc_list_push_cast(result->parts, (xml_dtc_object *) part);

    node = node->next;
  }

  return (xml_dtc_object *)result;
}

static int xml_dtc_aggregate_module_register(void) {
  xml_dtc_register_initializer("aggregator", (xml_dtc_handler_callback) xml_dtc_handler_aggregator_callback, NULL);
  xml_dtc_register_handler("aggregate", "pipeline", (xml_dtc_handler_callback) xml_dtc_handler_aggregate_callback, NULL);
  xml_dtc_register_type("aggregator", sizeof(xml_dtc_aggregator), NULL, "component");
  xml_dtc_register_type("aggregate-part", sizeof(xml_dtc_aggregate_part), (xml_dtc_object_free_callback) _xml_dtc_aggregate_part_free, NULL);
  xml_dtc_register_type("aggregate-caller", sizeof(xml_dtc_aggregator_caller), (xml_dtc_object_free_callback) _xml_dtc_aggregator_caller_free, "caller");
  return 0;
}

xml_dtc_module xml_dtc_aggregate_module = {
  XML_DTC_MODULE_STANDARD_STUFF,
  xml_dtc_aggregate_module_register
};
