/*
 * This module provides pipelines to the XML-DTC framework
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This module is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this module; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include <modules.h>
#include <xml-dtc.h>
#include <string.h>
#include <libxml/tree.h>

typedef struct _xml_dtc_pipeline xml_dtc_pipeline;

struct _xml_dtc_pipeline {
  XML_DTC_CALLER_STANDARD_DEF;
  xml_dtc_list *list;
};

static int _xml_dtc_pipeline_free(xml_dtc_pipeline *pipeline) {
  if (pipeline->list)
    xml_dtc_list_free(pipeline->list);
  return 0;
}

static xml_dtc_pipeline *xml_dtc_handler_pipeline_loop_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xml_dtc_pipeline *result;
  xml_dtc_hash *good_handlers;

  /* TODO: fix the memleak when returning NULL */

  good_handlers = (xml_dtc_hash *) xml_dtc_hash_get(handlers, (char *) node->name);

  result = (xml_dtc_pipeline *)xml_dtc_object_new(sitemap->resource, "pipeline", "pipeline");

  result->list = xml_dtc_object_list_new(sitemap->resource);

  node = node->children;

  while (node) {
     /* Ignore Blank Nodes */
    if (! xmlIsBlankNode(node)) {
      /* Test if node is an element or something else (PI, Comment...) */
      if (node->type != XML_ELEMENT_NODE) {
        /* Comments are allowed */
        if (node->type == XML_COMMENT_NODE) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Found comment");
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Element or comment expected");
          return NULL;
        }
      } else {
        /* Check if namespace is good */
        if ((node->ns != node->parent->ns) && (xmlStrcmp(node->ns->href, XML_DTC_SITEMAP_NAMESPACE) != 0)) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "expecting %s namespace, but got %s", XML_DTC_SITEMAP_NAMESPACE,
                      (node->ns != NULL) ? (char *) node->ns->href : "none");
          return NULL;
        }
        /* Now we know that namespace is ok, we can check and initialize all stuff */
        /* Search a handler for this node */
        {
          xml_dtc_handler *tmp;
          xml_dtc_object *object;
          if ((tmp = (xml_dtc_handler *)xml_dtc_hash_get(good_handlers, (char *)node->name))) {
            xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Found handler for %s", node->name);
            if ((object = tmp->handle(tmp, handlers, sitemap, node)) == NULL)
              return NULL;
            else
              xml_dtc_list_push_cast(result->list, object);
          } else {
            xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "No handler for %s", node->name);
            return NULL;
          }
        }
      }
    }
    node = node->next;
  }

  return result;
}

static xml_dtc_hash *xml_dtc_handler_pipelines_loop_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  /*  xml_dtc_container *result;*/
  xml_dtc_hash *good_handlers;

  /* TODO: fix the memleak when returning NULL */

  good_handlers = (xml_dtc_hash *) xml_dtc_hash_get(handlers, (char *) node->name);

  /*  result = xml_dtc_container_new(sitemap, (char *) node->name);*/

  node = node->children;

  while (node) {
    /* Ignore Blank Nodes */
    if (! xmlIsBlankNode(node)) {
      /* Test if node is an element or something else (PI, Comment...) */
      if (node->type != XML_ELEMENT_NODE) {
        /* Comments are allowed */
        if (node->type == XML_COMMENT_NODE) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Found comment");
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Element or comment expected");
          return NULL;
        }
      } else {
        /* Check if namespace is good */
        if ((node->ns != node->parent->ns) && (xmlStrcmp(node->ns->href, XML_DTC_SITEMAP_NAMESPACE) != 0)) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "expecting %s namespace, but got %s", XML_DTC_SITEMAP_NAMESPACE,
                      (node->ns != NULL) ? (char *) node->ns->href : "none");
          return NULL;
        }
        /* Now we know that namespace is ok, we can check and initialize all stuff */
        /* Search a handler for this node */
        {
          xml_dtc_handler *tmp;
          xml_dtc_object *object;
          if ((tmp = (xml_dtc_handler *)xml_dtc_hash_get(good_handlers, (char *)node->name))) {
            xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Found handler for %s", node->name);
            if ((object = tmp->handle(tmp, handlers, sitemap, node)) == NULL)
              return NULL;
            else
              xml_dtc_sitemap_add(sitemap, object);
          } else {
            xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "No handler for %s", node->name);
            return NULL;
          }
        }
      }
    }
    node = node->next;
  }

  /*  return result;*/
  return (xml_dtc_hash *)handler;
}

static int xml_dtc_run_pipelines(xml_dtc_request *request) {
  xml_dtc_list *pipelines;
  int i;
  pipelines = (xml_dtc_list *)xml_dtc_sitemap_get_all_type(request->sitemap, "pipeline");

  for (i = 0; (request->error.type == NULL) && (request->result == NULL) && (i < pipelines->len); i++) {
    xml_dtc_msg(request->sitemap->resource, XML_DTC_MSG_DEBUG, "Running pipeline #%d", i);
    xml_dtc_request_apply(request, ((xml_dtc_pipeline *)pipelines->list[i])->list);
  }

  if ((request->result == NULL) && (request->error.type == 0)) {
    xml_dtc_request_set_error(request, (xml_dtc_object *)request->sitemap, "resource-not-found", "Resource not found", "No pipeline matched the requested URI.");
  }
  return 0;
}

static int xml_dtc_pipelines_module_register(void) {
  xml_dtc_register_handler("pipelines", "sitemap", (xml_dtc_handler_callback) xml_dtc_handler_pipelines_loop_callback, NULL);
  xml_dtc_register_handler("pipeline", "pipelines", (xml_dtc_handler_callback) xml_dtc_handler_pipeline_loop_callback, NULL);
  xml_dtc_register_runner("pipeline", (xml_dtc_runner_callback) xml_dtc_run_pipelines, NULL);
  xml_dtc_register_type("pipeline", sizeof(xml_dtc_pipeline), (xml_dtc_object_free_callback) _xml_dtc_pipeline_free, "caller");

  return 0;
}

xml_dtc_module xml_dtc_pipelines_module = {
  XML_DTC_MODULE_STANDARD_STUFF,
  xml_dtc_pipelines_module_register
};
