/*
 * This module provides generation to the XML-DTC framework
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This module is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this module; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include <modules.h>
#include <xml-dtc.h>
#include <mappings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <libxml/HTMLtree.h>

typedef struct _xml_dtc_generator_caller xml_dtc_generator_caller;

typedef struct _xml_dtc_generator xml_dtc_generator;
struct _xml_dtc_generator {
  XML_DTC_COMPONENT_STANDARD_DEF;
};

struct _xml_dtc_generator_caller {
  XML_DTC_CALLER_STANDARD_DEF;
  xmlChar *src;
};

static xml_dtc_object *xml_dtc_handler_generator_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xmlChar *attr;
  xml_dtc_generator *result;

  attr = xmlGetProp(node, "name");
  result = (xml_dtc_generator *)xml_dtc_object_new(sitemap->resource, attr, (char *)node->name);
  result->handle = handler->_private;
  free(attr);

  xml_dtc_sitemap_add(sitemap, (xml_dtc_object *) result);
  return (xml_dtc_object *)handler;
}

static int _xml_dtc_generator_caller_free(xml_dtc_generator_caller *caller) {
  if (caller->src)
    xmlFree(caller->src);
  return 0;
}

static xml_dtc_object *xml_dtc_handler_generate_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xml_dtc_generator_caller *result;

  result = (xml_dtc_generator_caller *)xml_dtc_object_new(sitemap->resource, "generator", "generate-caller");
  result->component.name = xmlGetProp(node, "type");
  result->src = xmlGetProp(node, "src");

  node = node->children;

  while (node) {
    /* Ignore Blank Nodes */
    if (! xmlIsBlankNode(node)) {
      /* Test if node is an element or something else (PI, Comment...) */
      if (node->type != XML_ELEMENT_NODE) {
        /* Comments are allowed */
        if (node->type == XML_COMMENT_NODE) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Found comment");
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Element or comment expected");
          return NULL;
        }
      } else {
        /* Check if namespace is good */
        if ((node->ns != node->parent->ns) && (xmlStrcmp(node->ns->href, XML_DTC_SITEMAP_NAMESPACE) != 0)) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "expecting %s namespace, but got %s", XML_DTC_SITEMAP_NAMESPACE,
                      (node->ns != NULL) ? (char *) node->ns->href : "none");
          return NULL;
        }
        /* Now we know that namespace is ok, we can check and initialize all stuff */
        /* Search a handler for this node */
        {
	  if (xmlStrcmp(node->name, "parameter") == 0) {
	    xmlChar *tmp1, *tmp2;
	    if (!result->params) {
	      result->params = xml_dtc_string_hash_new(result->resource);
	    }
	    xml_dtc_string_hash_add(result->params, tmp1 = xmlGetProp(node, "name"), tmp2 = xmlGetProp(node, "value"));
	    xmlFree(tmp1);
	    xmlFree(tmp2);
	  }
        }
      }
    }
    node = node->next;
  }
  
  return (xml_dtc_object *)result;
}

static int xml_dtc_generate_file(xml_dtc_generator_caller *caller, xml_dtc_request *request) {
  xmlDocPtr xmldoc = NULL;
  char *filename;

  filename = xml_dtc_resolve_mappings(request->mappings, caller->src);
  xml_dtc_sitemap_resolve_relative_path(request->sitemap, &filename);

  xmldoc = xmlParseFile(filename);

  free(filename);

  if (xmldoc == NULL) {
    xml_dtc_msg(caller->resource, XML_DTC_MSG_ERROR, "File '%s' not parsed successfully.", caller->src);
    xml_dtc_request_set_error(request, (xml_dtc_object *)caller->component.object, "xml-error", "XML Parse Error", "XML Parse Error.");
    return -1;
  }

  request->xmltree = xmldoc;

  return 0;
}

static int xml_dtc_generate_html(xml_dtc_generator_caller *caller, xml_dtc_request *request) {
  xmlDocPtr xmldoc = NULL;
  char *filename;

  filename = xml_dtc_resolve_mappings(request->mappings, caller->src);
  xml_dtc_sitemap_resolve_relative_path(request->sitemap, &filename);

  xmldoc = htmlParseFile(filename, NULL);

  free(filename);

  if (xmldoc == NULL) {
    xml_dtc_msg(caller->resource, XML_DTC_MSG_ERROR, "File '%s' not parsed successfully.", caller->src);
    xml_dtc_request_set_error(request, (xml_dtc_object *)caller->component.object, "xml-error", "XML Parse Error", "XML Parse Error.");
    return -1;
  }

  request->xmltree = xmldoc;

  return 0;
}

static xmlNodePtr read_directory(xml_dtc_resource *resource, xmlNodePtr node, char *directory, int depth) {
  DIR *dir;
  dir = opendir(directory);

  if (dir) {
    xmlNodePtr current_node;
    struct dirent *dir_ent;
    char *file, size[255];
    struct stat stats;
    
    if (node == NULL) {
      xmlDocPtr xmldoc;
      xmlNsPtr ns;

      xmldoc = xmlNewDoc("1.0");
    
      if (xmldoc == NULL) {
	xml_dtc_msg(resource, XML_DTC_MSG_ERROR, "Could not create XML.");
	return NULL;
      }
    
      ns = xmlNewNs(NULL, "http://apache.org/cocoon/directory/2.0", "dir");
    
      node = xmlNewDocNode(xmldoc, ns, "directory", NULL);
      xmlDocSetRootElement(xmldoc, node);
      node->nsDef=ns;
      stat(directory, &stats);
      sprintf(size, "%d", (int) stats.st_size);
      xmlNewProp(node, BAD_CAST "size", size);
      sprintf(size, "%lu000", stats.st_mtime);
      xmlNewProp(node, BAD_CAST "lastModified", size);
    }
    
    while((dir_ent = readdir(dir))) {
      if (! ((dir_ent->d_name[0] == '.') && ((dir_ent->d_name[1] == 0) || ((dir_ent->d_name[1] == '.') && (dir_ent->d_name[2] == 0))))) {
	if (dir_ent->d_type == DT_REG) {
	  current_node = xmlNewChild(node, node->ns, BAD_CAST "file", NULL);
	} else {
	  current_node = xmlNewChild(node, node->ns, BAD_CAST "directory", NULL);
	}
	file = malloc(strlen(dir_ent->d_name) + strlen(directory) + 2);
	memcpy(file, directory, strlen(directory));
	file[strlen(directory)] = '/';
	memcpy(&file[strlen(directory) + 1], dir_ent->d_name, strlen(dir_ent->d_name) + 1);
	stat(file, &stats);
	xmlNewProp(current_node, BAD_CAST "name", dir_ent->d_name);
	sprintf(size, "%d", (int) stats.st_size);
	xmlNewProp(current_node, BAD_CAST "size", size);
	sprintf(size, "%lu000", stats.st_mtime);
	xmlNewProp(current_node, BAD_CAST "lastModified", size);
	if ((dir_ent->d_type != DT_REG) && (depth > 1)) {
	  read_directory(resource, current_node, file, depth - 1);
	}
	free(file);
      }
    }
    
    closedir(dir);
    return node;
  }
  return NULL;
}

static int xml_dtc_generate_directory(xml_dtc_generator_caller *caller, xml_dtc_request *request) {
  xmlNodePtr node;
  char *base_dir;
  int depth = 1;

  base_dir = xml_dtc_resolve_mappings(request->mappings, caller->src);
  xml_dtc_sitemap_resolve_relative_path(request->sitemap, &base_dir);

  if (caller->params) {
    char *depth_s;
    depth_s = xml_dtc_resolve_mappings(request->mappings, xml_dtc_hash_get(caller->params, "depth"));
    if (depth_s) {
      depth = atoi(depth_s);
    }
    free(depth_s);
  }
  node = read_directory(caller->resource, NULL, base_dir, depth);
  if (node == NULL) {
    xml_dtc_request_set_error(request, (xml_dtc_object *)caller->component.object, "resource-not-found", "Resource not found", "Directory cannot be opened.");
  } else {
    request->xmltree = node->doc;
  }

  free(base_dir);

  return 0;
}

static int xml_dtc_generate_request(xml_dtc_generator_caller *caller, xml_dtc_request *request) {
  xmlDocPtr xmldoc;
  xmlNodePtr root_node, node;
  xmlNsPtr ns;
  xml_dtc_hash *hash;
  char *src;
  int i;
  
  xmldoc = xmlNewDoc("1.0");

  if (xmldoc == NULL) {
    xml_dtc_msg(caller->resource, XML_DTC_MSG_ERROR, "Could not create XML.");
    return -1;
  }

  ns = xmlNewNs(NULL, "http://xml.apache.org/cocoon/requestgenerator/2.0", NULL);

  root_node = xmlNewDocNode(xmldoc, ns, "request", NULL);
  xmlDocSetRootElement(xmldoc, root_node);
  root_node->nsDef=ns;

  xmlNewProp(root_node, "target", request->uri);

  src = xml_dtc_resolve_mappings(request->mappings, caller->src);
  if (src)
    xmlNewProp(root_node, "source", src);

  hash = xml_dtc_request_get_headers(request);

  root_node = xmlNewChild(root_node, ns, BAD_CAST "requestHeaders", NULL);

  if (hash)
    for (i = 0; i < hash->num; i++) {
      node = xmlNewChild(root_node, ns, BAD_CAST "header", hash->list[i].element);
      
      xmlNewProp(node, "name", hash->list[i].key);
    }

  root_node = root_node->parent;

  hash = xml_dtc_request_get_parameters(request);

  root_node = xmlNewChild(root_node, ns, BAD_CAST "requestParameters", NULL);

  if (hash)
    for (i = 0; i < hash->num; i++) {
      node = xmlNewChild(root_node, ns, BAD_CAST "parameter", NULL);
      
      xmlNewProp(node, "name", hash->list[i].key);
      
      xmlNewChild(node, ns, BAD_CAST "value", hash->list[i].element);
    }
  
  root_node = root_node->parent;

  if (caller->params) {
    char *attr;
    attr = xml_dtc_hash_get(caller->params, "generate-attributes");
    if (attr && strncmp(attr, "true", 5) == 0) {
      hash = xml_dtc_request_get_attributes(request);

      root_node = xmlNewChild(root_node, ns, BAD_CAST "requestAttributes", NULL);
      if (hash)
	for (i = 0; i < hash->num; i++) {
	  node = xmlNewChild(root_node, ns, BAD_CAST "attribute", NULL);
	  
	  xmlNewProp(node, "name", hash->list[i].key);
	  
	  xmlNewChild(node, ns, BAD_CAST "value", hash->list[i].element);
	}
      
      root_node = root_node->parent;
    }

  }

  root_node = xmlNewChild(root_node, ns, BAD_CAST "configurationParameters", NULL);

  if ((hash = caller->params))
    for (i = 0; i < hash->num; i++) {
      node = xmlNewChild(root_node, ns, BAD_CAST "parameter", xml_dtc_resolve_mappings(request->mappings, hash->list[i].element));
      
      xmlNewProp(node, "name", hash->list[i].key);
    }
  
  request->xmltree = xmldoc;

  return 0;
}

static int xml_dtc_generators_register(void) {
  xml_dtc_register_handler("generators", "components", (xml_dtc_handler_callback) xml_dtc_handler_components_loop_callback, NULL);
  xml_dtc_register_handler("generate", "pipeline", (xml_dtc_handler_callback) xml_dtc_handler_generate_callback, NULL);
  xml_dtc_register_handler("file", "generators", (xml_dtc_handler_callback) xml_dtc_handler_generator_callback, xml_dtc_generate_file);
  xml_dtc_register_handler("directory", "generators", (xml_dtc_handler_callback) xml_dtc_handler_generator_callback, xml_dtc_generate_directory);
  xml_dtc_register_handler("request", "generators", (xml_dtc_handler_callback) xml_dtc_handler_generator_callback, xml_dtc_generate_request);
  xml_dtc_register_handler("html", "generators", (xml_dtc_handler_callback) xml_dtc_handler_generator_callback, xml_dtc_generate_html);
  xml_dtc_register_type("generator", sizeof(xml_dtc_generator), NULL, "component");
  xml_dtc_register_type("generate-caller", sizeof(xml_dtc_generator_caller), (xml_dtc_object_free_callback) _xml_dtc_generator_caller_free, "caller");

  return 0;
}

xml_dtc_module xml_dtc_generators_module = {
  XML_DTC_MODULE_STANDARD_STUFF,
  xml_dtc_generators_register
};
