/*
 * This module provides matching to the XML-DTC framework
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This module is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this module; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include <modules.h>
#include <xml-dtc.h>
#include <mappings.h>
#include <list.h>
#include <string.h>
#include <pcre.h>

typedef struct _xml_dtc_matcher_caller xml_dtc_matcher_caller;

typedef struct _xml_dtc_matcher xml_dtc_matcher;
struct _xml_dtc_matcher {
  XML_DTC_COMPONENT_STANDARD_DEF;
};

struct _xml_dtc_matcher_caller {
  XML_DTC_CALLER_STANDARD_DEF;
  xmlChar *pattern;
  xml_dtc_list *sub_pipeline;
  pcre *re;
};

static int _xml_dtc_matcher_free(xml_dtc_matcher *matcher) {
  if (matcher->_private)
    xmlFree(matcher->_private);
  return 0;}

static int _xml_dtc_matcher_caller_free(xml_dtc_matcher_caller *caller) {
  if (caller->sub_pipeline)
    xml_dtc_list_free(caller->sub_pipeline);
  if (caller->pattern)
    xmlFree(caller->pattern);
  if (caller->re)
    pcre_free(caller->re);
  return 0;
}

static xml_dtc_object *xml_dtc_handler_matcher_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xmlChar *attr;
  xml_dtc_matcher *result;

  attr = xmlGetProp(node, "name");
  result = (xml_dtc_matcher *)xml_dtc_object_new(sitemap->resource, attr, (char *)node->name);
  result->handle = handler->_private;
  free(attr);

  xml_dtc_sitemap_add(sitemap, (xml_dtc_object *) result);
  return (xml_dtc_object *)handler;
}

static xml_dtc_object *xml_dtc_handler_matcher_wildcard_header_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xmlChar *attr;
  xml_dtc_matcher *result;

  attr = xmlGetProp(node, "name");
  result = (xml_dtc_matcher *)xml_dtc_object_new(sitemap->resource, attr, (char *)node->name);
  result->handle = handler->_private;
  free(attr);

  node = node->children;

  while (node) {
    /* Ignore Blank Nodes */
    if (! xmlIsBlankNode(node)) {
      /* Test if node is an element or something else (PI, Comment...) */
      if (node->type != XML_ELEMENT_NODE) {
        /* Comments are allowed */
        if (node->type == XML_COMMENT_NODE) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Found comment");
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Element or comment expected");
          return NULL;
        }
      } else {
        /* There must be NO namespace */
        if (node->ns) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "expecting no namespace, but got %s",
                      node->ns->href);
          return NULL;
        }
        if (xmlStrcmp(node->name, "header-name") == 0) {
          result->_private = xmlNodeGetContent(node);
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "Unauthorized configuration %s", node->name);
          return NULL;
        }
      }
    }
    node = node->next;
  }

  xml_dtc_sitemap_add(sitemap, (xml_dtc_object *) result);
  return (xml_dtc_object *)handler;
}

static xml_dtc_object *xml_dtc_handler_matcher_wildcard_request_parameter_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xmlChar *attr;
  xml_dtc_matcher *result;

  attr = xmlGetProp(node, "name");
  result = (xml_dtc_matcher *)xml_dtc_object_new(sitemap->resource, attr, (char *)node->name);
  result->handle = handler->_private;
  free(attr);

  node = node->children;

  while (node) {
    /* Ignore Blank Nodes */
    if (! xmlIsBlankNode(node)) {
      /* Test if node is an element or something else (PI, Comment...) */
      if (node->type != XML_ELEMENT_NODE) {
        /* Comments are allowed */
        if (node->type == XML_COMMENT_NODE) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Found comment");
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Element or comment expected");
          return NULL;
        }
      } else {
        /* There must be NO namespace */
        if (node->ns) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "expecting no namespace, but got %s",
                      node->ns->href);
          return NULL;
        }
        if (xmlStrcmp(node->name, "parameter-name") == 0) {
          result->_private = xmlNodeGetContent(node);
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "Unauthorized configuration %s", node->name);
          return NULL;
        }
      }
    }
    node = node->next;
  }

  xml_dtc_sitemap_add(sitemap, (xml_dtc_object *) result);
  return (xml_dtc_object *)handler;
}

static xml_dtc_list *xml_dtc_handler_pipeline_loop_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xml_dtc_list *result;
  xml_dtc_hash *good_handlers;

  /* TODO: fix the memleak when returning NULL */

  good_handlers = (xml_dtc_hash *) xml_dtc_hash_get(handlers, (char *) node->name);

  result = xml_dtc_object_list_new(sitemap->resource);

  node = node->children;

  while (node) {
    /* Ignore Blank Nodes */
    if (! xmlIsBlankNode(node)) {
      /* Test if node is an element or something else (PI, Comment...) */
      if (node->type != XML_ELEMENT_NODE) {
        /* Comments are allowed */
        if (node->type == XML_COMMENT_NODE) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Found comment");
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Element or comment expected");
          return NULL;
        }
      } else {
        /* Check if namespace is good */
        if ((node->ns != node->parent->ns) && (xmlStrcmp(node->ns->href, XML_DTC_SITEMAP_NAMESPACE) != 0)) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "expecting %s namespace, but got %s", XML_DTC_SITEMAP_NAMESPACE,
                      (node->ns != NULL) ? (char *) node->ns->href : "none");
          return NULL;
        }
        /* Now we know that namespace is ok, we can check and initialize all stuff */
        /* Search a handler for this node */
        {
          xml_dtc_handler *tmp;
          xml_dtc_object *object;
          if ((tmp = (xml_dtc_handler *)xml_dtc_hash_get(good_handlers, (char *)node->name))) {
            xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Found handler for %s", node->name);
            if ((object = tmp->handle(tmp, handlers, sitemap, node)) == NULL)
              return NULL;
            else
              xml_dtc_list_push_cast(result, object);
          } else {
            xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "No handler for %s", node->name);
            return NULL;
          }
        }
      }
    }
    node = node->next;
  }

  return result;
}

static xml_dtc_object *xml_dtc_handler_match_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xml_dtc_matcher_caller *result;

  result = (xml_dtc_matcher_caller *)xml_dtc_object_new(sitemap->resource, "matcher", "match-caller");
  result->component.name = xmlGetProp(node, "type");
  result->pattern = xmlGetProp(node, "pattern");

  /* DIRTY HACK */
  xmlNodeSetName(node, "pipeline");
  result->sub_pipeline = xml_dtc_handler_pipeline_loop_callback(handler, handlers, sitemap, node);

  return (xml_dtc_object *)result;
}

static pcre *pattern2re(char *pattern) {
  int i, j, single = 0, dble = 0, esc = 0;
  const char *error;
  int erroffset;
  char *tmp;
  pcre *re;
  /* Note that there WILL be a problem if original pattern contains parenthesis */
  for (i = 0; pattern[i] != 0; i++) {
    if (pattern[i] == '*') {
      if (pattern[i + 1] == '*') {
	dble++;
	i++;
      } else {
	single++;
      }
    } else if ((pattern[i] == '+') || (pattern[i] == '.')) {
      esc++;
    }
  }
  tmp = malloc(i + single * 6 + dble * 3 + esc + 4);
  tmp[0] = '^';
  for (i = 0, j = 1; pattern[i] != 0; i++) {
    if (pattern[i] == '*')
      if (pattern[i + 1] == '*') {
	i++;
	tmp[j++] = '(';
	tmp[j++] = '.';
	tmp[j++] = '*';
	tmp[j++] = ')';
      } else {
	tmp[j++] = '(';
	tmp[j++] = '[';
	tmp[j++] = '^';
	tmp[j++] = '/';
	tmp[j++] = ']';
	tmp[j++] = '*';
	tmp[j++] = ')';
      }
    else if ((pattern[i] == '+') || (pattern[i] == '.')) {
      tmp[j++] = '\\';
      tmp[j++] = pattern[i];
    }
    else
      tmp[j++] = pattern[i];
  }
  tmp[j++] = '$';
  tmp[j] = 0;
  re = pcre_compile(tmp, PCRE_UTF8, &error, &erroffset, NULL);
  free(tmp);
  return re;
}

static int xml_dtc_match(xml_dtc_matcher_caller *caller, xml_dtc_request *request) {
  int rc;
  int ovector[30];

  if (!caller->re) {
    caller->re = pattern2re(caller->pattern);
  }

  rc = pcre_exec(caller->re, NULL, request->path_info, xmlStrlen(request->path_info), 0, 0, ovector, 30);
  if (rc > 0) {
    int i; xml_dtc_list *list; const char *result;
    list = xml_dtc_mappings_new_list(request->mappings);

    for (i = 0; i < rc; i++) {
      pcre_get_substring(request->path_info, ovector, rc, i, &result);
      xml_dtc_list_push_cast(list, (char *)result);
      free((char *)result);
    }
    i = xml_dtc_request_apply(request, caller->sub_pipeline);
    xml_dtc_mappings_destroy_last(request->mappings);
    return i;
  } else
    return 0;
}

static int xml_dtc_match_wildcard_header(xml_dtc_matcher_caller *caller, xml_dtc_request *request) {
  int rc;
  int ovector[30];
  char *header;
  xml_dtc_matcher *matcher;

  matcher = (xml_dtc_matcher *) caller->component.object;
  if (!matcher->_private)
    return -1;
  header = xml_dtc_request_get_header(request, matcher->_private);
  if (! header)
    return -1;

  if (!caller->re) {
    caller->re = pattern2re(caller->pattern);
  }

  rc = pcre_exec(caller->re, NULL, header, xmlStrlen(header), 0, 0, ovector, 30);

  if (rc > 0) {
    int i; xml_dtc_list *list; const char *result;
    list = xml_dtc_mappings_new_list(request->mappings);

    for (i = 0; i < rc; i++) {
      pcre_get_substring(header, ovector, rc, i, &result);
      xml_dtc_list_push_cast(list, (char *)result);
      free((char *)result);
    }

    i = xml_dtc_request_apply(request, caller->sub_pipeline);
    xml_dtc_mappings_destroy_last(request->mappings);
    return i;
  } else
    return 0;
}

static int xml_dtc_match_header(xml_dtc_matcher_caller *caller, xml_dtc_request *request) {
  int i;
  char *header;
  xml_dtc_list *list;

  header = xml_dtc_request_get_header(request, caller->pattern);
  if (! header)
    return -1;

  list = xml_dtc_mappings_new_list(request->mappings);

  xml_dtc_list_push_cast(list, header);
  xml_dtc_list_push_cast(list, header);

  i = xml_dtc_request_apply(request, caller->sub_pipeline);
  xml_dtc_mappings_destroy_last(request->mappings);
  return i;
}

static int xml_dtc_match_cookie(xml_dtc_matcher_caller *caller, xml_dtc_request *request) {
  int i;
  char *cookie;
  xml_dtc_list *list;

  cookie = xml_dtc_request_get_cookie(request, caller->pattern);
  if (! cookie)
    return -1;

  list = xml_dtc_mappings_new_list(request->mappings);

  xml_dtc_list_push_cast(list, cookie);
  xml_dtc_list_push_cast(list, cookie);

  i = xml_dtc_request_apply(request, caller->sub_pipeline);
  xml_dtc_mappings_destroy_last(request->mappings);
  return i;
}

static int xml_dtc_match_parameter(xml_dtc_matcher_caller *caller, xml_dtc_request *request) {
  int rc;
  int ovector[30];
  char *parameter;
  xml_dtc_matcher *matcher;

  matcher = (xml_dtc_matcher *) caller->component.object;
  if (!matcher->_private)
    return -1;
  parameter = xml_dtc_request_get_parameter(request, matcher->_private);
  if (! parameter)
    return -1;

  if (!caller->re) {
    caller->re = pattern2re(caller->pattern);
  }

  rc = pcre_exec(caller->re, NULL, parameter, xmlStrlen(parameter), 0, 0, ovector, 30);

  if (rc > 0) {
    int i; xml_dtc_list *list; const char *result;
    list = xml_dtc_mappings_new_list(request->mappings);

    for (i = 0; i < rc; i++) {
      pcre_get_substring(parameter, ovector, rc, i, &result);
      xml_dtc_list_push_cast(list, (char *)result);
      free((char *)result);
    }

    i = xml_dtc_request_apply(request, caller->sub_pipeline);
    xml_dtc_mappings_destroy_last(request->mappings);
    return i;
  } else
    return 0;
}

static int xml_dtc_match_request_parameter(xml_dtc_matcher_caller *caller, xml_dtc_request *request) {
  xml_dtc_list *list;
  char *param;

  param = xml_dtc_request_get_parameter(request, caller->pattern);
  if (param) {
    int i;
    list = xml_dtc_mappings_new_list(request->mappings);
    xml_dtc_list_push_cast(list, param);
    xml_dtc_list_push_cast(list, param);
    i = xml_dtc_request_apply(request, caller->sub_pipeline);
    xml_dtc_mappings_destroy_last(request->mappings);
    return i;
  } else
    return 0;
}

static int xml_dtc_matchers_module_register(void) {
  xml_dtc_register_handler("matchers", "components", (xml_dtc_handler_callback) xml_dtc_handler_components_loop_callback, NULL);
  xml_dtc_register_handler("match", "pipeline", (xml_dtc_handler_callback) xml_dtc_handler_match_callback, NULL);
  xml_dtc_register_handler("wildcard_uri", "matchers", (xml_dtc_handler_callback) xml_dtc_handler_matcher_callback, xml_dtc_match);
  xml_dtc_register_handler("wildcard_header", "matchers", (xml_dtc_handler_callback) xml_dtc_handler_matcher_wildcard_header_callback, xml_dtc_match_wildcard_header);
  xml_dtc_register_handler("request_parameter", "matchers", (xml_dtc_handler_callback) xml_dtc_handler_matcher_callback, xml_dtc_match_request_parameter);
  xml_dtc_register_handler("wildcard_request_parameter", "matchers", (xml_dtc_handler_callback) xml_dtc_handler_matcher_wildcard_request_parameter_callback, xml_dtc_match_parameter);
  xml_dtc_register_handler("header", "matchers", (xml_dtc_handler_callback) xml_dtc_handler_matcher_callback, xml_dtc_match_header);
  xml_dtc_register_handler("cookie", "matchers", (xml_dtc_handler_callback) xml_dtc_handler_matcher_callback, xml_dtc_match_cookie);
  xml_dtc_register_type("matcher", sizeof(xml_dtc_matcher), (xml_dtc_object_free_callback) _xml_dtc_matcher_free, "component");
  xml_dtc_register_type("match-caller", sizeof(xml_dtc_matcher_caller), (xml_dtc_object_free_callback) _xml_dtc_matcher_caller_free, "caller");
  return 0;
}

xml_dtc_module xml_dtc_matchers_module = {
  XML_DTC_MODULE_STANDARD_STUFF,
  xml_dtc_matchers_module_register
};
