/*
 * This module provides readers to the XML-DTC framework
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This module is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this module; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include <modules.h>
#include <xml-dtc.h>
#include <mappings.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

typedef struct _xml_dtc_reader_caller xml_dtc_reader_caller;

typedef struct _xml_dtc_reader xml_dtc_reader;
struct _xml_dtc_reader {
  XML_DTC_COMPONENT_STANDARD_DEF;
};

struct _xml_dtc_reader_caller {
  XML_DTC_CALLER_STANDARD_DEF;
  xmlChar *src, *mime_type;
};

static xml_dtc_object *xml_dtc_handler_reader_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xmlChar *attr;
  xml_dtc_reader *result;

  attr = xmlGetProp(node, "name");
  result = (xml_dtc_reader *)xml_dtc_object_new(sitemap->resource, attr, (char *)node->name);
  result->handle = handler->_private;
  free(attr);

  xml_dtc_sitemap_add(sitemap, (xml_dtc_object *) result);
  return (xml_dtc_object *)handler;
}

int _xml_dtc_reader_caller_free(xml_dtc_reader_caller *caller) {
  if (caller->src)
    xmlFree(caller->src);
  if (caller->mime_type)
    xmlFree(caller->mime_type);
  return 0;
}

static xml_dtc_object *xml_dtc_handler_read_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xml_dtc_reader_caller *result;
  result = (xml_dtc_reader_caller *)xml_dtc_object_new(sitemap->resource, "reader", "read-caller");
  result->component.name = xmlGetProp(node, "type");
  result->src = xmlGetProp(node, "src");
  result->mime_type = xmlGetProp(node, "mime-type");
  
  return (xml_dtc_object *)result;
}

static int xml_dtc_read_file(xml_dtc_reader_caller *caller, xml_dtc_request *request) {
  int fd;
  char *buf, *filename;
  struct stat file_stat;

  filename = xml_dtc_resolve_mappings(request->mappings, caller->src);
  xml_dtc_sitemap_resolve_relative_path(request->sitemap, &filename);

  fd = open(filename, O_RDONLY);
  free(filename);
  if (fd != -1) {
    if (fstat(fd, &file_stat) == 0) {
      if (S_ISDIR(file_stat.st_mode)) {
	xml_dtc_request_set_error(request, (xml_dtc_object *)caller->component.object, "resource-not-found", "Resource not found", "Cannot open file.");
	return -1;
      }
#if 0
      //      if (request->write) {
	int r;
	request->set_mime_type(request->write_context, caller->mime_type, NULL);
	request->set_content_length(request->write_context, file_stat.st_size);
	request->send_http_header(request->write_context);
	buf = malloc(409600);
	while ((r = read(fd, buf, 409600))) {
	  request->write(request->write_context, buf, r);
	}
	request->content_type = null;
	//} else {
#endif
	buf = malloc(file_stat.st_size);
	read(fd, buf, file_stat.st_size);
	request->result = buf;
	request->content_length = file_stat.st_size;
	request->mime_type = caller->mime_type;
	//      }
    }
    close(fd);
  } else {
    xml_dtc_request_set_error(request, (xml_dtc_object *)caller->component.object, "resource-not-found", "Resource not found", "Cannot open file.");
  }

  return 0;
}

static int xml_dtc_readers_module_register(void) {
  xml_dtc_register_handler("readers", "components", (xml_dtc_handler_callback) xml_dtc_handler_components_loop_callback, NULL);
  xml_dtc_register_handler("read", "pipeline", (xml_dtc_handler_callback) xml_dtc_handler_read_callback, NULL);
  xml_dtc_register_handler("resource", "readers", (xml_dtc_handler_callback) xml_dtc_handler_reader_callback, xml_dtc_read_file);
  xml_dtc_register_type("reader", sizeof(xml_dtc_reader), NULL, "component");
  xml_dtc_register_type("read-caller", sizeof(xml_dtc_reader_caller), (xml_dtc_object_free_callback) _xml_dtc_reader_caller_free, "caller");

  return 0;
}

xml_dtc_module xml_dtc_readers_module = {
  XML_DTC_MODULE_STANDARD_STUFF,
  xml_dtc_readers_module_register
};
