/*
 * This module provides serialization to the XML-DTC framework
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This module is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this module; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include <modules.h>
#include <xml-dtc.h>

typedef struct _xml_dtc_serializer_caller xml_dtc_serializer_caller;

typedef struct _xml_dtc_serializer xml_dtc_serializer;
struct _xml_dtc_serializer {
  XML_DTC_COMPONENT_STANDARD_DEF;
  xmlChar *doctype_public, *doctype_system, *encoding, *mime_type;
};

struct _xml_dtc_serializer_caller {
  XML_DTC_CALLER_STANDARD_DEF;
};

static int _xml_dtc_serializer_free(xml_dtc_serializer *serializer) {
  if (serializer->doctype_public)
    xmlFree(serializer->doctype_public);
  if (serializer->doctype_system)
    xmlFree(serializer->doctype_system);
  if (serializer->encoding)
    xmlFree(serializer->encoding);
  if (serializer->mime_type)
    xmlFree(serializer->mime_type);
  return 0;
}

static xml_dtc_object *xml_dtc_handler_serializer_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xmlChar *attr;
  xml_dtc_serializer *result;

  attr = xmlGetProp(node, "name");
  result = (xml_dtc_serializer *)xml_dtc_object_new(sitemap->resource, attr, (char *)node->name);
  result->handle = handler->_private;
  free(attr);
  result->mime_type = xmlGetProp(node, "mime-type");

  node = node->children;

  while (node) {
    /* Ignore Blank Nodes */
    if (! xmlIsBlankNode(node)) {
      /* Test if node is an element or something else (PI, Comment...) */
      if (node->type != XML_ELEMENT_NODE) {
        /* Comments are allowed */
        if (node->type == XML_COMMENT_NODE) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Found comment");
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Element or comment expected");
          return NULL;
        }
      } else {
        /* There must be NO namespace */
        if (node->ns) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "expecting no namespace, but got %s",
                      node->ns->href);
          return NULL;
        }
        if (xmlStrcmp(node->name, "doctype-public") == 0) {
          result->doctype_public = xmlNodeGetContent(node);
        } else if (xmlStrcmp(node->name, "doctype-system") == 0) {
          result->doctype_system = xmlNodeGetContent(node);
        } else if (xmlStrcmp(node->name, "encoding") == 0) {
          result->encoding = xmlNodeGetContent(node);
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "Unauthorized configuration %s", node->name);
          return NULL;
        }
      }
    }
    node = node->next;
  }



  xml_dtc_sitemap_add(sitemap, (xml_dtc_object *) result);
  return (xml_dtc_object *)handler;
}

static xml_dtc_object *xml_dtc_handler_serialize_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xml_dtc_serializer_caller *result;

  result = (xml_dtc_serializer_caller *)xml_dtc_object_new(sitemap->resource, "serializer", "serialize-caller");
  result->component.name = xmlGetProp(node, "type");

  return (xml_dtc_object *)result;
}

static int xml_dtc_serialize(xml_dtc_serializer_caller *caller, xml_dtc_request *request) {
  if (! request->internal) {
    int size;
    xmlNodePtr node;
    xml_dtc_serializer *serializer = (xml_dtc_serializer *) caller->component.object;
    node = xmlDocGetRootElement(request->xmltree);

    if (serializer->doctype_public) {
      xmlCreateIntSubset(request->xmltree, node->name, serializer->doctype_public, serializer->doctype_system);
    }
    if (serializer->encoding)
      xmlDocDumpMemoryEnc(request->xmltree, &request->result, &size, serializer->encoding);
    else
      xmlDocDumpMemory(request->xmltree, &request->result, &size);
    xmlFreeDoc(request->xmltree);
    request->xmltree = NULL;

    request->content_length = size;
    request->mime_type = serializer->mime_type;
    request->encoding = serializer->encoding;

    //xmlCleanupParser();

    //xmlMemoryDump();
  }
  return 0;
}

static int xml_dtc_serializers_module_register(void) {
  xml_dtc_register_handler("serializers", "components", (xml_dtc_handler_callback) xml_dtc_handler_components_loop_callback, NULL);
  xml_dtc_register_handler("serialize", "pipeline", (xml_dtc_handler_callback) xml_dtc_handler_serialize_callback, NULL);
  xml_dtc_register_handler("xml", "serializers", (xml_dtc_handler_callback) xml_dtc_handler_serializer_callback, xml_dtc_serialize);
  xml_dtc_register_type("serializer", sizeof(xml_dtc_serializer), (xml_dtc_object_free_callback) _xml_dtc_serializer_free, "component");
  xml_dtc_register_type("serialize-caller", sizeof(xml_dtc_serializer_caller), NULL, "caller");
  return 0;
}

xml_dtc_module xml_dtc_serializers_module = {
  XML_DTC_MODULE_STANDARD_STUFF,
  xml_dtc_serializers_module_register
};
