/*
 * This module provides transformation to the XML-DTC framework
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This module is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this module; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include <modules.h>
#include <xml-dtc.h>
#include <mappings.h>
#include <string.h>
#include <libxslt/xslt.h>
#include <libxslt/xsltInternals.h>
#include <libxslt/transform.h>
#ifdef WITH_LIBEXSLT
#include <libexslt/exslt.h>
#endif
#include <config.h>

#if 0
static xml_dtc_hash *xml_dtc_xslt_cache = NULL;
#endif

typedef struct _xml_dtc_transformer_caller xml_dtc_transformer_caller;

typedef struct _xml_dtc_transformer xml_dtc_transformer;
struct _xml_dtc_transformer {
  XML_DTC_COMPONENT_STANDARD_DEF;
};

struct _xml_dtc_transformer_caller {
  XML_DTC_CALLER_STANDARD_DEF;
  xmlChar *src;
};

static int _xml_dtc_transformer_caller_free(xml_dtc_transformer_caller *caller) {
  if (caller->src)
    xmlFree(caller->src);
  return 0;
}

static xml_dtc_object *xml_dtc_handler_transformer_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xmlChar *attr;
  xml_dtc_transformer *result;

  attr = xmlGetProp(node, "name");
  result = (xml_dtc_transformer *)xml_dtc_object_new(sitemap->resource, attr, (char *)node->name);
  result->handle = handler->_private;
  xmlFree(attr);

  xml_dtc_sitemap_add(sitemap, (xml_dtc_object *) result);
  return (xml_dtc_object *)handler;
}

static xml_dtc_object *xml_dtc_handler_transform_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xml_dtc_transformer_caller *result;

  result = (xml_dtc_transformer_caller *)xml_dtc_object_new(sitemap->resource, "transformer", "transform-caller");
  result->component.name = xmlGetProp(node, "type");
  result->src = xmlGetProp(node, "src");

  node = node->children;

  while (node) {
    /* Ignore Blank Nodes */
    if (! xmlIsBlankNode(node)) {
      /* Test if node is an element or something else (PI, Comment...) */
      if (node->type != XML_ELEMENT_NODE) {
        /* Comments are allowed */
        if (node->type == XML_COMMENT_NODE) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Found comment");
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Element or comment expected");
          return NULL;
        }
      } else {
        /* Check if namespace is good */
        if ((node->ns != node->parent->ns) && (xmlStrcmp(node->ns->href, XML_DTC_SITEMAP_NAMESPACE) != 0)) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "expecting %s namespace, but got %s", XML_DTC_SITEMAP_NAMESPACE,
                      (node->ns != NULL) ? (char *) node->ns->href : "none");
          return NULL;
        }
        /* Now we know that namespace is ok, we can check and initialize all stuff */
        /* Search a handler for this node */
        {
	  if (xmlStrcmp(node->name, "parameter") == 0) {
	    xmlChar *tmp1, *tmp2;
	    if (!result->params) {
	      result->params = xml_dtc_string_hash_new(result->resource);
	    }
	    xml_dtc_string_hash_add(result->params, tmp1 = xmlGetProp(node, "name"), tmp2 = xmlGetProp(node, "value"));
	    xmlFree(tmp1);
	    xmlFree(tmp2);
	  }
        }
      }
    }
    node = node->next;
  }

  return (xml_dtc_object *)result;
}

static const char *_xsltparams[] = { "GENERATOR_NAME", "'" PACKAGE_NAME "'", "GENERATOR_VERSION", "'" PACKAGE_VERSION "'", "GENERATOR_STRING", "'" PACKAGE_STRING "'", "GENERATOR_HOMEPAGE", "'" PACKAGE_HOMEPAGE "'", "GENERATOR_PREFIX", "'" PACKAGE_NAME "'", "PATH_INFO", NULL, "REQUEST_URI", NULL, NULL };

static int xml_dtc_transform(xml_dtc_transformer_caller *caller, xml_dtc_request *request) {
  xsltStylesheetPtr xslt = NULL;
  xmlDocPtr doc;
  char **xsltparams;
  char *filename;
  int i;

  xsltparams = malloc(sizeof(char *) * (15 + (caller->params?(caller->params->num * 2):0)));

  for(i = 0; i < 15; i++)
    xsltparams[i] = (char *)_xsltparams[i];
  xsltparams[11] = malloc(strlen(request->path_info) + 3);
  xsltparams[11][0] = '\'';
  memcpy(&xsltparams[11][1], request->path_info, strlen(request->path_info));
  xsltparams[11][strlen(request->path_info)+1] = '\'';
  xsltparams[11][strlen(request->path_info)+2] = 0;
  xsltparams[13] = malloc(strlen(request->uri) + 3);
  xsltparams[13][0] = '\'';
  memcpy(&xsltparams[13][1], request->uri, strlen(request->uri));
  xsltparams[13][strlen(request->uri)+1] = '\'';
  xsltparams[13][strlen(request->uri)+2] = 0;

  /* BEURK HACK */
  if (caller->params) {
    int j;
    for (j = 0, i = 0; j < caller->params->num; j++)
      if (caller->params->list[j].key) {
	char *tmp = xml_dtc_resolve_mappings(request->mappings, caller->params->list[j].element);
	xsltparams[++i + 13] = caller->params->list[j].key;
	xsltparams[++i + 13] = malloc(strlen(tmp) + 3);
	xsltparams[i + 13][0] = '\'';
	memcpy(&xsltparams[i + 13][1], tmp, strlen(tmp));
	xsltparams[i + 13][strlen(tmp) + 1] = '\'';
	xsltparams[i + 13][strlen(tmp) + 2] = 0;
	free(tmp);
      }
    xsltparams[++i + 13] = NULL;
  }

  filename = xml_dtc_resolve_mappings(request->mappings, caller->src);
  xml_dtc_sitemap_resolve_relative_path(request->sitemap, &filename);

#if 0
  if (xml_dtc_xslt_cache) {
    xslt = (xsltStylesheetPtr) xml_dtc_hash_get(xml_dtc_xslt_cache, filename);
  } else {
    xml_dtc_xslt_cache = xml_dtc_hash_new(request->resource, "xslt cache");
  }
  if (xslt == NULL) {
#endif
    xslt = xsltParseStylesheetFile(filename);
#if 0
    xml_dtc_hash_add(xml_dtc_xslt_cache, strdup(filename), (char *)xslt);
  }
#endif

  if (xslt == NULL) {
    xml_dtc_msg(caller->resource, XML_DTC_MSG_ERROR, "File '%s' not parsed successfully.", caller->src);
    xml_dtc_request_set_error(request, (xml_dtc_object *)caller->component.object, "xslt-error", "XSLT Parse Error", "XSLT Parse Error.");
    return -1;
  }

  doc = xsltApplyStylesheet(xslt, request->xmltree, (const char **) xsltparams);
  xsltFreeStylesheet(xslt);
  free(filename);
  i = 10;
  while(xsltparams[i++]) {
    free(xsltparams[i++]);
  }
  free(xsltparams);
  if (doc == NULL) {
    xml_dtc_msg(caller->resource, XML_DTC_MSG_ERROR, "Error during XSLT processing.", caller->src);
    xml_dtc_request_set_error(request, (xml_dtc_object *)caller->component.object, "xslt-error", "XSLT Transformation Error", "XSLT Transformation Error.");
    return -1;
  }
  xmlFreeDoc(request->xmltree);
  request->xmltree = doc;
  return 0;
}

static int xml_dtc_transformers_module_register(void) {
  xml_dtc_register_handler("transformers", "components", (xml_dtc_handler_callback) xml_dtc_handler_components_loop_callback, NULL);
  xml_dtc_register_handler("transform", "pipeline", (xml_dtc_handler_callback) xml_dtc_handler_transform_callback, NULL);
  xml_dtc_register_handler("xslt", "transformers", (xml_dtc_handler_callback) xml_dtc_handler_transformer_callback, xml_dtc_transform);
  xml_dtc_register_type("transformer", sizeof(xml_dtc_transformer), NULL, "component");
  xml_dtc_register_type("transform-caller", sizeof(xml_dtc_transformer_caller), (xml_dtc_object_free_callback) _xml_dtc_transformer_caller_free, "caller");
#ifdef WITH_LIBEXSLT
  exsltRegisterAll();
#endif
  return 0;
}

static void xml_dtc_transformers_module_close(void) {
#if 0
  // BEWARE WHEN IT WILL BE REINTRODUCED (because of the use of hashes) !
  if (xml_dtc_xslt_cache) {
    int i;
    for (i = 0; i < xml_dtc_xslt_cache->num; i++) {
      xsltFreeStylesheet((xsltStylesheetPtr) xml_dtc_xslt_cache->values[i]);
      xml_dtc_xslt_cache->values[i] = 0;
    }
    xml_dtc_object_unlink(xml_dtc_xslt_cache);
  }
#endif
  xsltCleanupGlobals();
}

xml_dtc_module xml_dtc_transformers_module = {
  XML_DTC_MODULE_STANDARD_STUFF,
  xml_dtc_transformers_module_register,
  xml_dtc_transformers_module_close
};
